#define VC_EXTRALEAN
#include <windows.h>

// Macros //////////////////////////////////////////////////////////////////////

// General Macros:

#define DWORD_ALIGN(x) ((x) + 3 & ~3)

// The longest unformatted error message in english is probably
// ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE_V2 (368 characters including the
// terminating null) I don't know if more than 512 characters are required in
// other languages.
#define MAXSTR_TEXT 512

#define MAXSTR_UI64 21

// Configuration-Specific Macros:

#ifndef _DEBUG
#define DEBUG_ONLY(x) __noop
#else // #ifndef _DEBUG
#define DEBUG_ONLY(x) x
#endif // #ifndef _DEBUG

#ifndef UNICODE
#define CF_PTSTR CF_TEXT
#define TCHAR_DEC(r) dec r
#define TCHAR_PTR byte ptr
#define TCHAR_REG(r) r##l
#else // #ifndef UNICODE
#define CF_PTSTR CF_UNICODETEXT
#define TCHAR_DEC(r) sub r, type TCHAR
#define TCHAR_PTR word ptr
#define TCHAR_REG(r) r##x
#endif // #ifndef UNICODE

#if defined(_M_IX86) && !defined(NO_ASM)
#define USE_ASM
#endif // #if defined(_M_IX86) && !defined(NO_ASM)

// Constants ///////////////////////////////////////////////////////////////////

static const LPVOID _lpSource =
TEXT("High-resolution performance counter frequency: %1 Hz.");
static const LPCTSTR _lpCaption =
TEXT("QPFTest");

// Data Type Declarations //////////////////////////////////////////////////////

typedef union
{
    LARGE_INTEGER li;
    TCHAR sz[MAXSTR_UI64];
}
Frequency;

// Function Declarations ///////////////////////////////////////////////////////

static LPTSTR
__stdcall UI64ToString(const ULONGLONG *pll, TCHAR szBuffer[MAXSTR_UI64]);

// Function Definitions ////////////////////////////////////////////////////////

#ifdef USE_ASM
void __declspec(naked) __stdcall EntryPoint()
{

    /***********************************************************************

    The following registers have constant values throughout the function:
    ebx is 0, NULL, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT).
    esi is the address of a memory block on the stack that is used as a
    HANDLE before label m1 and as a buffer for the output message
    (MAXSTR_TEXT TCHAR's) after label m1.
    edi is FormatMessage

    ***********************************************************************/

    __asm
    {
        ; ebx, esi and edi don't need to be saved since this is the program
        ; entry point. On Windows 9x ebp must be saved.
        push ebp
        mov ebp, esp
        sub esp, DWORD_ALIGN(1 * type LPVOID) + type Frequency
        mov eax, esp
        sub esp, DWORD_ALIGN(MAXSTR_TEXT * type TCHAR)
        mov esi, esp

        xor ebx, ebx
        mov edi, dword ptr FormatMessage

        ; eax is the address of a memory block on the stack that is used to
        ; hold the frequency both as a LARGE_INTEGER and as a null-
        ; terminated TCHAR string. This is possible since UI64ToString
        ; correctly overwrites an output region that overlaps with the input
        ; data.
        push eax        ; szBuffer in UI64ToString
        push eax        ; pll in UI64ToString

        push eax        ; lpFrequency
        call dword ptr QueryPerformanceFrequency
        test eax, eax
        je m2

        call UI64ToString
        mov [ebp - 4], eax

        push eax        ; Arguments: ignored
        push ebx        ; nSize: 0
        push esi        ; lpBuffer: address of a handle on the stack
        push eax        ; dwLanguageId: ignored
        push eax        ; dwMessageId: ignored
        push eax        ; lpSource: returned in eax by UI64ToString
        push dword ptr \
            FORMAT_MESSAGE_ALLOCATE_BUFFER | \
            FORMAT_MESSAGE_IGNORE_INSERTS | \
            FORMAT_MESSAGE_FROM_STRING
        call edi        ; FormatMessage
        test eax, eax
        je m1

        push ebx        ; hWndNewOwner: NULL
        call dword ptr OpenClipboard
        call dword ptr EmptyClipboard
        push [esi]      ; hMem
        push dword ptr CF_PTSTR
        call dword ptr SetClipboardData
        call dword ptr CloseClipboard

m1:
        mov edx, FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ARGUMENT_ARRAY
        push dword ptr MB_ICONINFORMATION
        jmp m3

m2:
        call dword ptr GetLastError
        mov edx, FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS
        push dword ptr MB_ICONERROR

m3:
        lea ecx, [ebp - 4]
        push ecx        ; Arguments
        push dword ptr MAXSTR_TEXT
        push esi        ; lpBuffer: allocated on the stack
        push ebx        ; MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT)
        push eax        ; dwMessageId: ignored if not returned by GetLastError
        push _lpSource  ; lpSource
        push edx        ; dwFlags
        call edi        ; FormatMessage
        test eax, eax
        je m4           ; return

        push _lpCaption ; lpCaption
        push esi        ; lpText
        push ebx        ; hWnd: NULL
        call dword ptr MessageBox

m4:
        leave
        ret 0
    }
}
#else // #ifdef USE_ASM
#pragma warning(push)
#pragma warning(disable: 4701)
void __stdcall EntryPoint()
{
    void *Arguments[1];
    DWORD dwFlags;
    DWORD dwMessageId;
    Frequency frequency;
    UINT uType;

    if (QueryPerformanceFrequency(&frequency.li))
    {
        Arguments[0] =
            UI64ToString((ULONGLONG *)&frequency.li.QuadPart, frequency.sz);

        {
            HANDLE hFrequency;
            DWORD dwFormatMessage =
                FormatMessage(
                FORMAT_MESSAGE_ALLOCATE_BUFFER |
                FORMAT_MESSAGE_IGNORE_INSERTS |
                FORMAT_MESSAGE_FROM_STRING,
                Arguments[0],
                0,
                0,
                (LPTSTR)&hFrequency,
                0,
                NULL
                );
            if (dwFormatMessage != 0)
            {
                OpenClipboard(NULL);
                EmptyClipboard();
                SetClipboardData(CF_PTSTR, hFrequency);
                CloseClipboard();
            }
        }

        dwFlags = FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ARGUMENT_ARRAY;
        // The following assignment prevents a run-time error that could
        // otherwise occur in a debug build compiled with the /RTCu option.
        DEBUG_ONLY(dwMessageId = 0);
        uType = MB_ICONINFORMATION;
    }
    else
    {
        dwFlags = FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS;
        dwMessageId = GetLastError();
        uType = MB_ICONERROR;
    }

    {
        TCHAR szText[MAXSTR_TEXT];
        DWORD dwFormatMessage =
            FormatMessage(
            dwFlags,
            _lpSource,
            dwMessageId,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            szText,
            MAXSTR_TEXT,
            (va_list *)Arguments
            );
        if (dwFormatMessage != 0) MessageBox(NULL, szText, _lpCaption, uType);
    }
}
#pragma warning(pop)
#endif // #ifdef USE_ASM

#ifdef USE_ASM
LPTSTR __declspec(naked)
__stdcall UI64ToString(const ULONGLONG *pll, TCHAR szBuffer[MAXSTR_UI64])
{
    pll; szBuffer; // unused parameters

    __asm
    {
        push ebx
        push esi
        push edi

        mov ebx, [esp + 16] ; szBuffer
        add ebx, (MAXSTR_UI64 - 1) * type TCHAR
        mov eax, [esp + 20] ; pll
        mov esi, [eax + 4]
        mov edi, [eax]
        mov TCHAR_PTR [ebx], '\0'
        mov ecx, 10

m1:
        TCHAR_DEC(ebx)
        xor edx, edx
        mov eax, esi
        div ecx
        mov esi, eax
        mov eax, edi
        div ecx
        mov edi, eax

        add edx, '0'
        mov [ebx], TCHAR_REG(d)
        or eax, esi
        jnz m1

        mov eax, ebx

        pop edi
        pop esi
        pop ebx
        ret 8
    }
}
#else // #ifdef USE_ASM
LPTSTR __stdcall UI64ToString(const ULONGLONG *pll, TCHAR szBuffer[MAXSTR_UI64])
{
    ULONGLONG ll = *pll;
    TCHAR *pch = szBuffer + (MAXSTR_UI64 - 1);

    *pch = 0;
    do
    {
        *--pch = (CHAR)('0' + (ll % 10));
        ll /= 10;
    }
    while (ll != 0);
    return pch;
}
#endif // #ifdef USE_ASM
